package be.iminds.ilabt.jfed.experimenter_gui.canvas.impl;

import be.iminds.ilabt.jfed.experimenter_gui.canvas.CanvasLink;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.ExperimentCanvas;
import be.iminds.ilabt.jfed.rspec.model.GeantTestbedType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import javafx.beans.InvalidationListener;
import javafx.beans.WeakInvalidationListener;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.event.EventHandler;
import javafx.scene.control.Button;
import javafx.scene.input.MouseEvent;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Visualisation class for Rspec ifaceLinks.
 * <p>
 * These can have a connection to more than two nodes, and are thus a collection of
 * two OR MORE ifaceLinks to a central "link"-component
 * User: twalcari
 * Date: 11/13/13
 * Time: 9:22 AM
 */
public class GtsToRspecCanvasLink {

    private static final Logger LOG = LoggerFactory.getLogger(GtsToRspecCanvasLink.class);

    private static final String LINK_CENTER_STYLECLASS = "gts-link-center";

    private final String gtsPortName;
    private final GeantTestbedType gtt;
    private final FXRspecNode node;

    private final ExperimentCanvas experimentCanvas;
    private final LinkCenter linkCenter;
    private final CanvasLink toGttLink, toNodeLink;
    private final InvalidationListener layoutListener = observable -> updateLayout();

    private double centerX, centerY;
    private final ObjectProperty<EventHandler<MouseEvent>> onMouseClicked = new SimpleObjectProperty<>();
    private final RspecCanvasNode rspecCanvasNode;
    private final RspecGTSNode rspecGTSNode;

    public GtsToRspecCanvasLink(String gtsPortName, GeantTestbedType gtt, FXRspecNode node, ExperimentCanvas experimentCanvas) {
        this.gtsPortName = gtsPortName;
        this.gtt = gtt;
        this.node = node;

        this.experimentCanvas = experimentCanvas;

        rspecCanvasNode = experimentCanvas.getRspecCanvasNode(node);
        rspecGTSNode = experimentCanvas.getRspecGTSNode(gtt);

        //create centernode
        this.linkCenter = new LinkCenter();

        this.toNodeLink = new CanvasLink(rspecCanvasNode, linkCenter);
        this.toGttLink = new CanvasLink(rspecGTSNode, linkCenter);
        experimentCanvas.getCanvas().getChildren().add(0, linkCenter);
        experimentCanvas.getCanvas().getChildren().add(0, toGttLink);
        experimentCanvas.getCanvas().getChildren().add(0, toNodeLink);

        rspecCanvasNode.layoutXProperty().addListener(new WeakInvalidationListener(layoutListener));
        rspecCanvasNode.layoutYProperty().addListener(new WeakInvalidationListener(layoutListener));
        rspecCanvasNode.boundsInLocalProperty().addListener(new WeakInvalidationListener(layoutListener));

        rspecGTSNode.layoutXProperty().addListener(new WeakInvalidationListener(layoutListener));
        rspecGTSNode.layoutYProperty().addListener(new WeakInvalidationListener(layoutListener));
        rspecGTSNode.boundsInLocalProperty().addListener(new WeakInvalidationListener(layoutListener));

    }


    private void updateLayout() {
        centerX = (CanvasLink.getCenterX(rspecCanvasNode) + CanvasLink.getCenterX(rspecGTSNode)) / 2;
        centerY = (CanvasLink.getCenterY(rspecCanvasNode) + CanvasLink.getCenterY(rspecGTSNode)) / 2;

        CanvasLink.setCenterX(centerX, linkCenter);
        CanvasLink.setCenterY(centerY, linkCenter);
    }

    /**
     * Should be called when the RspecCanvasLink is removed from the canvas
     */
    public void remove() {
        experimentCanvas.getCanvas().getChildren().removeAll(linkCenter, toNodeLink, toGttLink);

    }


    public double getCenterX() {
        return centerX;
    }

    public double getCenterY() {
        return centerY;
    }


    public Button getLinkCenter() {
        return linkCenter;
    }


    public void setOnMouseClicked(EventHandler<MouseEvent> onMouseClicked) {
        this.onMouseClicked.set(onMouseClicked);
    }

    public EventHandler<MouseEvent> getOnMouseClicked() {
        return onMouseClicked.get();
    }

    public class LinkCenter extends Button {

        public LinkCenter() {
            super();
            setText("GTS: [" + gtsPortName + "]");

            getStyleClass().add(LINK_CENTER_STYLECLASS);
            setOnAction(actionEvent -> onLinkCenterClicked());
            layoutBoundsProperty().addListener(layoutListener);

            onMouseClickedProperty().bind(GtsToRspecCanvasLink.this.onMouseClicked);


        }

        private void onLinkCenterClicked() {
            experimentCanvas.getSelectionProvider().setSelectedObject(GtsToRspecCanvasLink.this);
        }

        public GtsToRspecCanvasLink getGtsToRspecCanvasLink() {
            return GtsToRspecCanvasLink.this;
        }
    }

    public String getGtsPortName() {
        return gtsPortName;
    }

    public GeantTestbedType getGtt() {
        return gtt;
    }

    public FXRspecNode getNode() {
        return node;
    }
}
