package be.iminds.jfed.gts_highlevel.controller;

import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.JFedPasswordManager;
import be.iminds.jfed.gts_highlevel.model.*;
import com.google.common.collect.Maps;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.FXCollections;
import javafx.collections.MapChangeListener;
import javafx.collections.ObservableMap;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Singleton;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 9/24/2015
 * Time: 15:48
 */
@Singleton
public class GtsModel {
    private static final Logger LOG = LoggerFactory.getLogger(GtsModel.class);
    private final JFedPasswordManager passwordManager;

    private final ObjectProperty<GtsCredentials> credentials = new SimpleObjectProperty<>();
    private final ObservableMap<String, GtsProject> projects = FXCollections.observableHashMap();
    private final ObservableMap<String, GtsInfrastructure> infrastructures = FXCollections.observableHashMap();
    private final ObservableMap<Integer, GtsReservation> reservations = FXCollections.observableHashMap();
    private final ObservableMap<String, GtsResource>  resources = FXCollections.observableHashMap();
    private final ObservableMap<TypeId, GtsType> types = FXCollections.observableHashMap();

    private final ObservableMap<Integer, TypeId> reservationsByTypeId = FXCollections.observableHashMap();

    @Inject
    public GtsModel(JFedPasswordManager passwordManager,
                    JFedPreferences jFedPreferences,
                    GeniUserProvider geniUserProvider) {
        this.passwordManager = passwordManager;

        passwordManager.load(jFedPreferences, geniUserProvider.getLoggedInGeniUser());

        resources.addListener(new MapChangeListener<String, GtsResource>() {
            @Override
            public void onChanged(Change<? extends String, ? extends GtsResource> change) {
                LOG.warn("Change to resources-map: {}", change);
            }
        });
    }


    public GtsCredentials getCredentials() {
        return credentials.get();
    }

    public ObjectProperty<GtsCredentials> credentialsProperty() {
        return credentials;
    }

    public void setCredentials(GtsCredentials credentials) {
        this.credentials.set(credentials);
    }

    public String getUsername() {
        return passwordManager.getLogins("gts").get(0).getUsername();
    }

    public ObservableMap<String, GtsProject> getProjects() {
        return /*FXCollections.unmodifiableObservableMap(*/projects/*)*/;
    }

    void addOrUpdateProject(GtsProject project) {
        projects.put(project.getId(), project);
    }

    /**
     * Register a project as removed
     */
    void removeProject(GtsProject project) {
        projects.remove(project.getId());
    }

    public ObservableMap<String, GtsInfrastructure> getInfrastructures() {
        return /*FXCollections.unmodifiableObservableMap(*/infrastructures/*)*/;
    }

    void addOrUpdateInfrastructure(GtsInfrastructure infrastructure) {
        infrastructures.put(infrastructure.getProviderId(), infrastructure);
    }

    /**
     * Register a infrastructure as removed
     */
    void removeInfrastructure(GtsInfrastructure infrastructure) {
        infrastructures.remove(infrastructure.getProviderId());
    }

    public ObservableMap<Integer, GtsReservation> getReservations() {
        return reservations;
    }


    GtsReservation addOrUpdateReservation(GtsReservation newReservation) {
        List<GtsReservation> newChildrenList = newReservation.getChildren().stream()
                .map(this::addOrUpdateReservation)
                .collect(Collectors.toList());

        newReservation.setChildren(newChildrenList);

        return reservations.merge(newReservation.getId(), newReservation, (existingReservation, newReservation1) -> {
            GtsReservation.copy(existingReservation, newReservation1);
            return existingReservation;
        });
    }

    /**
     * Register a reservation as removed
     */
    void removeReservation(GtsReservation reservation) {
        LOG.warn("Removing reservation {}", reservation.getId());
        reservations.remove(reservation.getId());
    }


    public ObservableMap<String, GtsResource> getResources() {
        return resources;
    }

    public Set<GtsResource> getResourcesInReservation(GtsReservation reservation) {
        assert reservations.containsValue(reservation);

        Set<GtsResource> result = new HashSet<>();
        getResourcesInReservation(reservation, result);
        return result;
    }

    private void getResourcesInReservation(GtsReservation reservation, Set<GtsResource> result) {
        if (resources.containsKey(reservation.getResourcesProviderId())) {
            result.add(resources.get(reservation.getResourcesProviderId()));
        }

        reservation.getChildren().forEach(childRes -> getResourcesInReservation(childRes, result));
    }


    void addOrUpdateResource(GtsResource newResource) {
        newResource.setReservations(addOrUpdateReservation(newResource.getReservations()));

        resources.merge(newResource.getProviderId(), newResource, (existingResource, newResource1) -> {
            GtsResource.copy(existingResource, newResource1);
            return existingResource;
        });

        LOG.debug("Saw resource {}. Model now has {} in total", newResource, resources.size());
    }

    /**
     * Register a resource as removed
     */
    void removeResource(GtsResource resource) {
        resources.remove(resource.getProviderId());
    }


    public ObservableMap<TypeId, GtsType> getTypes() {
        return /*FXCollections.unmodifiableObservableMap(*/types/*)*/;
    }

    public Map<TypeId, GtsType> getTypesInProject(GtsProject project) {
        return Maps.filterEntries(types, input -> input.getKey().projectName.equals(project.getId()));
    }


    void addOrUpdateType(GtsType type, GtsProject project) {
        types.put(new TypeId(project.getId(), type.getName()), type);
    }

    /**
     * Register a type as removed
     */
    boolean removeType(GtsType type, GtsProject project) {
        return types.remove(new TypeId(project.getId(), type.getName())) != null;
    }

    /**
     * Register a type as removed
     */
    boolean removeType(String typeName, GtsProject project) {
        return types.remove(new TypeId(project.getId(), typeName)) != null;
    }

    boolean removeProject(String projectName) {
        return projects.remove(projectName) != null;
    }

    boolean removeReservation(int reservationId) {
        return reservations.remove(reservationId) != null;
    }

    @Nullable
    public GtsReservation getReservation(int reservationId) {
        return reservations.get(reservationId);
    }


    public void registerReservationOfType(Integer reservationId, TypeId typeId){
        reservationsByTypeId.put(reservationId, typeId);
    }

    public void deregisterReservationOfType(Integer reservationId){
        reservationsByTypeId.remove(reservationId);
    }

    public ObservableMap<Integer, TypeId> getReservationsByTypeId() {
        return /*FXCollections.unmodifiableObservableMap(*/reservationsByTypeId/*)*/;
    }

    public final static class TypeId {
        private final String projectName;
        private final String typeName;

        public TypeId(@Nonnull String projectName, @Nonnull String typeName) {
            this.projectName = projectName;
            this.typeName = typeName;
        }

        public String getProjectName() {
            return projectName;
        }

        public String getTypeName() {
            return typeName;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            TypeId typeId = (TypeId) o;

            if (!projectName.equals(typeId.projectName)) return false;
            return typeName.equals(typeId.typeName);

        }

        @Override
        public int hashCode() {
            int result = projectName.hashCode();
            result = 31 * result + typeName.hashCode();
            return result;
        }

        @Override
        public String toString() {
            return "TypeId{" +
                    "projectName='" + projectName + '\'' +
                    ", typeName='" + typeName + '\'' +
                    '}';
        }
    }

}
