package be.iminds.jfed.gts_highlevel.tasks;

import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.util.JavaFXLogger;
import be.iminds.jfed.gts_highlevel.controller.GtsException;
import be.iminds.jfed.gts_highlevel.controller.GtsModel;
import be.iminds.jfed.gts_highlevel.controller.GtsReservationManager;
import be.iminds.jfed.gts_highlevel.controller.GtsReservationManagerFactory;
import be.iminds.jfed.gts_highlevel.model.GtsProject;
import be.iminds.jfed.gts_highlevel.model.GtsReservation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Provider;
import java.util.List;
import java.util.Optional;

/**
 * User: twalcari
 * Date: 10/7/2015
 * Time: 15:45
 */
public class UpdateGtsReservationsUntilReadyTask extends LoggedInGtsTask {
    private static final Logger LOG = LoggerFactory.getLogger(UpdateGtsReservationsUntilReadyTask.class);
    private static final int PERIOD_MS = 10_000;

    private final GtsProject project;
    private final int reservationId;
    private final TaskThread tt;

    private boolean finished = false;
    private boolean success = false;
    private int failCount = 0;
    private int triesCount = 0;
    private String status = "";


    /**
     * Create task that will periodically update the reservations list until the given reservation reaches a final state
     *
     * @param project       Project in which the reservation resides
     * @param reservationId ID of reservation that needs to be checked until final state is set
     */
    UpdateGtsReservationsUntilReadyTask(GtsProject project, int reservationId,
                                        GtsReservationManagerFactory gtsReservationManagerFactory,
                                        Provider<GtsLoginTask> loginGtsTaskProvider, GtsModel gtsModel,
                                        JavaFXLogger logger, TaskThread tt) {
        super("Update reservations in project '" + project.getId() + "' for '" + reservationId + "'", logger, gtsReservationManagerFactory, loginGtsTaskProvider, gtsModel);
        this.reservationId = reservationId;
        this.project = project;
        this.tt = tt;
    }

    @Override
    protected void doGtsRMTask(TaskExecution taskExecution, GtsReservationManager reservationManager) throws GtsException {
        triesCount++;

        List<GtsReservation> reservations = reservationManager.listReservations(project);

        Optional<GtsReservation> res =
                reservations.stream().filter(newRes -> newRes.getId() == reservationId).findFirst();

        if (res.isPresent()) {
            if (!isFinalState(res.get().getStatus())) {
                LOG.debug("Reservation {} is in transitioning state {}. Updating again in 10 seconds....", reservationId, res.get().getStatus());
                tt.scheduleTask(this, PERIOD_MS);
            } else {
                LOG.info("Reservation {} reached final state {}. Quiting updates.", reservationId, res.get().getStatus());
                finished = true;
                success = true;
            }
        } else {
            LOG.error("Could not find reservation {} in reservations list received! Quiting updates.", reservationId);
            this.status = "Could not find reservation. Aborting checks!";
            finished = true;
            failCount++;
        }
    }

    public long getMsUntilNextCheck() {
        if (!futureTasks.isEmpty()) {
            return futureTasks.get(0).getTimeLeftMs();
        } else
            return 0;

    }

    public boolean isFinished() {
        return finished;
    }

    private static boolean isFinalState(GtsReservation.Status status) {
        return status == GtsReservation.Status.ACTIVE ||
                status == GtsReservation.Status.DEACTIVED ||
                status == GtsReservation.Status.FAILURE;
    }

    public int getTriesCount() {
        return triesCount;
    }

    public String getStatus() {
        return status;
    }

    public boolean isSuccess() {
        return success;
    }

    public int getReservationId() {
        return reservationId;
    }
}
