package be.iminds.ilabt.jfed.experiment;

import be.iminds.ilabt.jfed.experiment.util.ExperimentRestoreInformation;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.GeniUrn;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.ReadOnlyObjectProperty;
import javafx.beans.property.ReadOnlyObjectWrapper;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.FXCollections;
import javafx.collections.MapChangeListener;
import javafx.collections.ObservableMap;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * User: twalcari
 * Date: 1/30/2015
 * Time: 13:24
 */
public class Experiment {
    private static final Logger LOG = LoggerFactory.getLogger(Experiment.class);

    /*
    Variables with basic information. These fields will always be set in the constructor
     */
    @Nonnull
    private final String name;
    @Nullable
    private final String subAuthName; //may be null

    @Nullable
    private final RequestRspecSource newRequestRspecSource;

    @Nullable
    private final RequestRspecSource existingRequestRspecSource;

    @Nullable
    /**
     * No sub-second precision allowed (FEDIBBTDEV-738). Must always be truncated to second precision!
     */
    private final Instant requestedStartTime;
    @Nullable
    /**
     * No sub-second precision allowed (FEDIBBTDEV-738). Must always be truncated to second precision!
     */
    private final Instant requestedEndTime;
    private final boolean includeProjectMemberSshKeys;
    private final boolean requestSla;

    private ExperimentRestoreInformation experimentRestoreInformation;

    /*
    Variables which are subject to change troughout the lifecycle of an experiment
     */
    private final ReadOnlyObjectWrapper<Slice> slice = new ReadOnlyObjectWrapper<>();

    private final ObjectProperty<ExperimentState> experimentState = new SimpleObjectProperty<>();

    private final ObservableMap<ExperimentPartAuthority, ExperimentPart> experimentParts =
            FXCollections.observableHashMap();

    /*
    Listeners
     */
    private final List<ExperimentChangeListener> experimentChangeListeners = new ArrayList<>();

    //instance initializer
    {
        this.experimentState.addListener((observable, oldValue, newValue) -> {
            experimentChangeListeners.forEach(ecl -> ecl.onExperimentStateChange(newValue));
        });

        this.experimentParts.addListener((MapChangeListener<ExperimentPartAuthority, ExperimentPart>) change -> {
            if (change.wasAdded()) {
                experimentChangeListeners.forEach(ecl -> ecl.onExperimentPartAdded(change.getKey(), change.getValueAdded()));
            } else {
                throw new IllegalStateException("ExperimentParts cannot be removed!");
            }
        });
    }


    /**
     * Constructor used when starting a new experiment in a new to-be-created slice
     */
    public Experiment(@Nonnull String name, @Nullable String subAuthName, @Nonnull RequestRspecSource newRequestRspecSource,
                      @Nullable Instant requestedStartTime, @Nonnull Instant requestedEndTime,
                      boolean includeProjectMemberSshKeys, boolean requestSla) {
        this.name = name;
        this.subAuthName = subAuthName;
        this.requestSla = requestSla;

        this.requestedStartTime =
                requestedStartTime != null ? requestedStartTime.truncatedTo(ChronoUnit.SECONDS) : null;
        this.requestedEndTime =
                requestedEndTime.truncatedTo(ChronoUnit.SECONDS); //No sub-second precision allowed (FEDIBBTDEV-738)

        this.newRequestRspecSource = newRequestRspecSource;
        this.existingRequestRspecSource = null;

        this.includeProjectMemberSshKeys = includeProjectMemberSshKeys;

        this.experimentState.setValue(ExperimentState.PENDING);
    }


    /**
     * Constructor used when starting a new experiment in an existing slice
     */
    public Experiment(@Nonnull Slice slice, @Nonnull RequestRspecSource newRequestRspecSource,
                      @Nullable Instant requestedStartTime, @Nonnull Instant requestedEndTime,
                      boolean includeProjectMemberSshKeys, @Nullable RequestRspecSource existingRequestRspecSource, boolean requestSla) {
        this.requestSla = requestSla;
        this.name = slice.getName();
        this.subAuthName = slice.getSubAuthorityName();

        this.requestedStartTime =
                requestedStartTime != null ? requestedStartTime.truncatedTo(ChronoUnit.SECONDS) : null;
        this.requestedEndTime =
                requestedEndTime.truncatedTo(ChronoUnit.SECONDS); //No sub-second precision allowed (FEDIBBTDEV-738)

        this.newRequestRspecSource = newRequestRspecSource;
        this.existingRequestRspecSource = existingRequestRspecSource;

        this.includeProjectMemberSshKeys = includeProjectMemberSshKeys;

        this.experimentState.setValue(ExperimentState.RESTORING);
        registerSlice(slice);
    }


    /**
     * Constructor used when restoring a slice.
     */
    public Experiment(@Nonnull Slice slice, @Nullable Instant requestedStartTime) {
        this.name = slice.getName();
        this.subAuthName = slice.getSubAuthorityName();


        this.newRequestRspecSource = null;

        RequestRspecSource existingRequestRspecSource = slice.getRequestRspec();
        if (existingRequestRspecSource == null && slice.getManifestRspec() != null) {
            existingRequestRspecSource =
                    new RequestRspecSource(slice.getManifestRspec().getStringRspec(), ModelRspecType.FX);
        }
        this.existingRequestRspecSource = existingRequestRspecSource;

        this.requestedStartTime = requestedStartTime; //to be fixed slice.getStartDate();
        this.requestedEndTime = null;


        this.includeProjectMemberSshKeys = false;
        this.requestSla = false;
        registerSlice(slice);

        this.experimentState.set(ExperimentState.RESTORING);

    }

    public void registerSlice(Slice slice) {
        if (this.slice.get() != null)
            throw new IllegalStateException("Slice is already set in Experiment.");
        this.slice.setValue(slice);

        this.experimentRestoreInformation = new ExperimentRestoreInformation(slice.getUrn());
        experimentRestoreInformation.bindToExperiment(this);
    }


    public void addExperimentChangeListener(ExperimentChangeListener experimentChangeListener) {
        assert (!this.experimentChangeListeners.contains(experimentChangeListener));

        this.experimentChangeListeners.add(experimentChangeListener);
    }

    public void removeExperimentChangeListener(ExperimentChangeListener experimentChangeListener) {
        assert (this.experimentChangeListeners.contains(experimentChangeListener));

        this.experimentChangeListeners.remove(experimentChangeListener);
    }


    @Nonnull
    public String getName() {
        return name;
    }

    public boolean isIncludeProjectMemberSshKeys() {
        return includeProjectMemberSshKeys;
    }

    public boolean isRequestSla() {
        return requestSla;
    }

    @Nullable
    public Instant getRequestedEndTime() {
        return requestedEndTime;
    }

    @Nullable
    public Instant getRequestedStartTime() {
        return requestedStartTime;
    }

    @Nullable
    public RequestRspecSource getNewRequestRspecSource() {
        return newRequestRspecSource;
    }

    @Nullable
    public Slice getSlice() {
        return slice.get();
    }

    public ReadOnlyObjectProperty<Slice> sliceProperty() {
        return slice.getReadOnlyProperty();
    }


    public ExperimentState getExperimentState() {
        return experimentState.get();
    }

    public ObjectProperty<ExperimentState> experimentStateProperty() {
        return experimentState;
    }

    @Nullable
    public String getSubAuthName() {
        return subAuthName;
    }

    public boolean isReservationMade() {
        return experimentParts.values().stream()
                .anyMatch(experimentPart -> experimentPart.getState() == InternalState.RESERVED);
    }

    public ExperimentRestoreInformation getExperimentRestoreInformation() {
        return experimentRestoreInformation;
    }

    public void setExperimentState(ExperimentState experimentState) {
        LOG.debug("Setting slice state for experiment {} to {}", name, experimentState);
        this.experimentState.set(experimentState);

    }

    @Nullable
    public RequestRspecSource getExistingRequestRspecSource() {
        return existingRequestRspecSource;
    }

    public ObservableMap<ExperimentPartAuthority, ExperimentPart> getParts() {
        return experimentParts;
    }

    public void addExperimentPart(ExperimentPart experimentPart) {
        assert experimentPart.getExperiment() == this;
        assert !experimentParts.containsKey(experimentPart.getAuthority());

        this.experimentParts.put(experimentPart.getAuthority(), experimentPart);
    }

    public void registerExperimentPart(ExperimentPart experimentPart) {
        if (!experimentParts.containsKey(experimentPart.getAuthority())) {
            this.experimentParts.put(experimentPart.getAuthority(), experimentPart);
        } else {
            throw new IllegalArgumentException("There is already an experiment part registered for authority " +
                    experimentPart.getAuthority());
        }
    }

    @Nullable
    public ExperimentPart getPart(SfaAuthority auth) {
        return experimentParts.entrySet().stream()
                .filter(entry -> entry.getKey().getSfaAuthority().equals(auth))
                .map(Map.Entry::getValue)
                .findFirst().orElse(null);
    }

    @Nullable
    public ExperimentPart getPart(Sliver sliver) {
        return experimentParts.values().stream()
                .filter(experimentPart -> experimentPart.getSlivers() != null &&
                        experimentPart.getSlivers().contains(sliver))
                .findFirst().orElse(null);
    }

    public ExperimentPart getPart(GeniUrn sliverUrn) {
        for (ExperimentPart experimentPart : experimentParts.values()) {
            if(experimentPart.getSlivers() == null)
                continue;

            for (Sliver sliver : experimentPart.getSlivers()) {
                if (sliver.getUrn().equals(sliverUrn))
                    return experimentPart;
            }
        }
        return null;
    }
}
