package be.iminds.ilabt.jfed.experiment;

import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.ReadOnlyObjectProperty;
import javafx.beans.property.ReadOnlyObjectWrapper;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.FXCollections;
import javafx.collections.ListChangeListener;
import javafx.collections.ObservableList;

import javax.annotation.Nullable;
import java.time.Instant;
import java.util.Collection;
import java.util.Comparator;
import java.util.Objects;
import java.util.stream.Collectors;

/**
 * An ExperimentPart is a collection of resources within an @link{Experiment} which are being managed by
 * one connection-authority.
 * <p/>
 * ex.
 * - all slivers managed by ExoGeni SM
 * - one GTS reservation of one GTS type
 */
//TODO add expiration time
public class ExperimentPart {
    private final Experiment experiment;
    private final ExperimentPartAuthority authority;

    private final ObjectProperty<InternalState> state = new SimpleObjectProperty<>();

    private final ObservableList<Sliver> slivers = FXCollections.observableArrayList();
    private final ReadOnlyObjectWrapper<Instant> expirationTime = new ReadOnlyObjectWrapper<>();

    public ExperimentPart(Experiment experiment, SfaAuthority authority) {
        this.experiment = experiment;
        assert experiment.getSlice() != null;
        assert experiment.getSlice().getSlivers().isEmpty() : "Expecting an empty slivers list!";

        this.authority = new ExperimentPartAuthority(authority);

        slivers.addListener((ListChangeListener<Sliver>) c -> {
            while (c.next()) {
                for (Sliver sliver : c.getAddedSubList()) {
                    sliver.expirationDateProperty().addListener(observable -> updateExpirationTime());
                }
            }
            updateExpirationTime();
        });


        experiment.getSlice().getSlivers().addListener(new ListChangeListener<Sliver>() {
            @Override
            public void onChanged(Change<? extends Sliver> c) {
                while (c.next()) {
                    //TODO: replace by AuthorityFinder ?
                    slivers.addAll(c.getAddedSubList().stream()
                            .filter(sliver -> sliver.getAuthority().getUrn().getEncodedTopLevelAuthority_withoutSubAuth().equals(
                                    authority.getUrn().getEncodedTopLevelAuthority_withoutSubAuth()))
                            .collect(Collectors.toList()));
                    c.getRemoved().forEach(slivers::remove);
                }
            }
        });


    }

    public Experiment getExperiment() {
        return experiment;
    }

    public ExperimentPartAuthority getAuthority() {
        return authority;
    }

    public String getName() {
        return getSfaAuthority().getName();
    }

    public SfaAuthority getSfaAuthority() {
        return authority.getSfaAuthority();
    }


    public InternalState getState() {
        return state.get();
    }

    public ObjectProperty<InternalState> stateProperty() {
        return state;
    }

    public void setState(InternalState state) {
        this.state.set(state);
    }

    @Nullable
    public Collection<Sliver> getSlivers() {
        if (experiment.getSlice() == null)
            return null;
        return experiment.getSlice().findSlivers(authority.getSfaAuthority());
    }

    public ReadOnlyObjectProperty<Instant> expirationTimeProperty() {
        return expirationTime.getReadOnlyProperty();
    }

    @Nullable
    public Instant getExpirationTime() {
        return expirationTime.get();
    }

    private void updateExpirationTime() {
        expirationTime.set(
                slivers.stream()
                        .map(Sliver::getExpirationDate)
                        .filter(Objects::nonNull)
                        .min(Comparator.naturalOrder()).orElse(null));
    }

}
