package be.iminds.ilabt.jfed.experiment.tasks;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import javafx.application.Platform;
import javafx.beans.property.ReadOnlyObjectProperty;
import javafx.beans.property.ReadOnlyObjectWrapper;
import javafx.concurrent.Task;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;

/**
 * User: twalcari
 * Date: 2/2/2015
 * Time: 10:29
 */
public abstract class ExperimentTask extends Task<ExperimentTaskStatus> {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentTask.class);

    protected final Experiment experiment;

    private final ReadOnlyObjectWrapper<ExperimentTaskStatus> status = new ReadOnlyObjectWrapper<>(ExperimentTaskStatus.INACTIVE);
    protected final List<TaskExecutionFinishedListener> taskExecutionFinishedListeners = new ArrayList<>();
    private final List<ExperimentTaskFinishedCallback> finishedCallbacks = new ArrayList<>();

    public ExperimentTask(Experiment experiment) {
        this.experiment = experiment;

    }

    @Override
    protected final ExperimentTaskStatus call() throws Exception {
        updateStatus(ExperimentTaskStatus.BUSY);

        ExperimentTaskStatus finalStatus = executeTask();

        assert (finalStatus == ExperimentTaskStatus.SUCCESS ||
                finalStatus == ExperimentTaskStatus.WARNING ||
                finalStatus == ExperimentTaskStatus.FAILED) : "Unexpected final status";

        updateStatus(finalStatus);

        for (ExperimentTaskFinishedCallback callback : finishedCallbacks)
            callback.onExperimentTaskFinished(finalStatus);

        return finalStatus;
    }

    private void updateStatus(final ExperimentTaskStatus newStatus) {
        Platform.runLater(() -> status.set(newStatus));
    }


    protected abstract ExperimentTaskStatus executeTask();

    public Experiment getExperiment() {
        return experiment;
    }

    public void addTaskFinishedCallback(ExperimentTaskFinishedCallback callback) {
        finishedCallbacks.add(callback);
    }

    protected <T extends be.iminds.ilabt.jfed.highlevel.controller.Task> void registerFinishedTaskExecution(TaskExecution<T> taskExecution) {
        taskExecutionFinishedListeners.forEach(listener -> listener.onTaskExecutionFinished(taskExecution));
    }

    public void addTaskExecutionsListener(TaskExecutionFinishedListener listener) {
        taskExecutionFinishedListeners.add(listener);
    }

    protected <T extends be.iminds.ilabt.jfed.highlevel.controller.Task> void waitForTaskExecutionToComplete(TaskExecution<T> taskExecution) {
        while (!taskExecution.isCompleted()) {
            try {
                Thread.sleep(100);
            } catch (InterruptedException ignored) {
            }
        }

        registerFinishedTaskExecution(taskExecution);
    }

    @Override
    protected void failed() {
        LOG.error("An uncaught exception caused ExperimentTask " + getClass().getSimpleName() + " to fail!", getException());

    }

    public ExperimentTaskStatus getStatus() {
        return status.get();
    }

    public ReadOnlyObjectProperty<ExperimentTaskStatus> statusProperty() {
        return status.getReadOnlyProperty();
    }
}
