package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.model.EasyModel;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecLink;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.util.GeniUrn;

import javax.annotation.Nonnull;
import javax.inject.Inject;
import javax.inject.Provider;
import javax.inject.Singleton;
import java.time.Instant;
import java.util.Collection;
import java.util.List;

@Singleton
public final class JobFactory {
    private final Provider<HighLevelTaskFactory> hltfProvider;
    private final Provider<TaskThread> ttProvider;
    private final Provider<AuthorityFinder> authorityFinderProvider;
    private final Provider<AuthorityListModel> authorityListModelProvider;
    private final Provider<AggregateManagerWrapperFactory> aggregateManagerWrapperFactoryProvider;
    private final Provider<GeniUserProvider> geniUserProviderProvider;
    private final Provider<EasyModel> easyModelProvider;
    private final Provider<JobStateFactory> jobStateFactoryProvider;


    @Inject
    public JobFactory(Provider<HighLevelTaskFactory> hltfProvider,
                      Provider<TaskThread> ttProvider,
                      Provider<AuthorityFinder> authorityFinderProvider,
                      Provider<AuthorityListModel> authorityListModelProvider,
                      Provider<AggregateManagerWrapperFactory> aggregateManagerWrapperFactoryProvider,
                      Provider<GeniUserProvider> geniUserProviderProvider,
                      Provider<EasyModel> easyModelProvider,
                      Provider<JobStateFactory> jobStateFactoryProvider) {
        this.hltfProvider = hltfProvider;
        this.ttProvider = ttProvider;
        this.authorityFinderProvider = authorityFinderProvider;
        this.authorityListModelProvider = authorityListModelProvider;
        this.aggregateManagerWrapperFactoryProvider = aggregateManagerWrapperFactoryProvider;
        this.geniUserProviderProvider = geniUserProviderProvider;
        this.easyModelProvider = easyModelProvider;
        this.jobStateFactoryProvider = jobStateFactoryProvider;
    }

    public StartExperimentJob createStartExperimentJob(Experiment experiment) {
        return new StartExperimentJob(experiment,
                hltfProvider.get(), ttProvider.get(), authorityFinderProvider.get(),
                aggregateManagerWrapperFactoryProvider.get(), geniUserProviderProvider.get(), easyModelProvider.get(), jobStateFactoryProvider.get());
    }

    public RestoreExperimentJob createRestoreExperimentJob(Experiment experiment) {
        return new RestoreExperimentJob(experiment, hltfProvider.get(), ttProvider.get(), authorityFinderProvider.get(), jobStateFactoryProvider.get());
    }

    public UpdateExperimentJob createUpdateExperimentJob(Experiment experiment) {
        return new UpdateExperimentJob(experiment, hltfProvider.get(), ttProvider.get());
    }

    public StopExperimentJob createStopExperimentJob(Experiment experiment) {
        return new StopExperimentJob(experiment, hltfProvider.get(), ttProvider.get(), jobStateFactoryProvider.get());
    }

    public StopExperimentJob createStopExperimentJob(Experiment experiment, Collection<ExperimentPart> experimentParts) {
        return new StopExperimentJob(experiment, experimentParts, hltfProvider.get(), ttProvider.get(), jobStateFactoryProvider.get());
    }

    public ShareSliceJob createShareSliceJob(Experiment experiment, Collection<GeniUrn> usersToShareWith, boolean registerSshKeys) {
        return new ShareSliceJob(experiment, usersToShareWith, registerSshKeys, hltfProvider.get(), ttProvider.get());
    }

    public EditSshKeysJob createEditSshKeysJob(Experiment experiment, List<UserSpec> userspec) {
        return new EditSshKeysJob(experiment, userspec, hltfProvider.get(), ttProvider.get(), aggregateManagerWrapperFactoryProvider.get(), jobStateFactoryProvider.get());
    }

    public ReloadOSJob createReloadOSJob(Experiment experiment, Sliver sliver) {
        return new ReloadOSJob(experiment, sliver, hltfProvider.get(), ttProvider.get(), jobStateFactoryProvider.get());
    }

    public CreateDiskImageJob createCreateDiskImageJob(Experiment experiment, FXRspecNode node, String imageName, boolean global, boolean updatePrepare) {
        return new CreateDiskImageJob(experiment, node, imageName, global, updatePrepare, hltfProvider.get(), ttProvider.get(), jobStateFactoryProvider.get());
    }

    public RegisterExperimentJob createRegisterExperimentJob(Experiment experiment) {
        return new RegisterExperimentJob(experiment, hltfProvider.get(), ttProvider.get());
    }

    public RebootJob createRebootJob(Experiment experiment, Sliver sliver) {
        return new RebootJob(experiment, sliver, hltfProvider.get(), ttProvider.get(), jobStateFactoryProvider.get());
    }

    public RebootJob createRebootJob(ExperimentPart experimentPart) {
        return new RebootJob(experimentPart, hltfProvider.get(), ttProvider.get(), jobStateFactoryProvider.get());
    }

    public OpenConsoleJob createOpenConsoleJob(Experiment experiment, Sliver sliver) {
        return new OpenConsoleJob(experiment, sliver, hltfProvider.get(), ttProvider.get());
    }


    public ShareLanJob createShareLanJob(Experiment experiment, FXRspecLink link, String sharedLanName) {
        return new ShareLanJob(experiment, link, sharedLanName, true /* share */, hltfProvider.get(), ttProvider.get(), authorityFinderProvider.get());
    }

    public ShareLanJob createUnshareLanJob(Experiment experiment, FXRspecLink link, String sharedLanName) {
        return new ShareLanJob(experiment, link, sharedLanName, false /* unshare */, hltfProvider.get(), ttProvider.get(), authorityFinderProvider.get());
    }

    public FetchSliceMemberSshKeysJob createFetchSliceMemberSshKeysJob(Experiment experiment) {
        return new FetchSliceMemberSshKeysJob(experiment, hltfProvider.get(), ttProvider.get());
    }

    public RenewJob createRenewJob(Experiment experiment, Instant newExpirationTime) {
        return new RenewJob(experiment, newExpirationTime, hltfProvider.get(), ttProvider.get());
    }

    public RenewJob createRenewJob(Experiment experiment, Instant newExpirationTime, Collection<ExperimentPart> experimentParts) {
        return new RenewJob(experiment, newExpirationTime, experimentParts, hltfProvider.get(), ttProvider.get());
    }

    public StartStitchedExperimentJob createStartStitchedExperimentJob(@Nonnull  Experiment experiment, @Nonnull  SfaAuthority scsAuthority) {
       return new StartStitchedExperimentJob(experiment, scsAuthority, hltfProvider.get(), ttProvider.get(),
               jobStateFactoryProvider.get(),  authorityListModelProvider.get(),
               authorityFinderProvider.get(), aggregateManagerWrapperFactoryProvider.get());
    }
}
