package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.controller.JFedHighLevelException;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.tasks.RenewSliceTask;
import be.iminds.ilabt.jfed.highlevel.tasks.RenewSliversTask;
import be.iminds.ilabt.jfed.lowlevel.JFedException;

import java.time.Instant;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Collection;
import java.util.List;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 12/2/2015
 * Time: 15:54
 */
public class RenewJob extends AbstractJob<ExperimentTaskStatus> {

    private final Instant newExpirationDate;
    private final Collection<ExperimentPart> experimentParts;

    public RenewJob(Experiment experiment, Instant newExpirationDate, HighLevelTaskFactory hltf, TaskThread tt) {
        super("Renew experiment until " + DateTimeFormatter.RFC_1123_DATE_TIME.format(ZonedDateTime.ofInstant(newExpirationDate, ZoneId.systemDefault())),
                experiment, hltf, tt);
        this.newExpirationDate = newExpirationDate;
        this.experimentParts = null;
    }

    public RenewJob(Experiment experiment, Instant newExpirationDate, Collection<ExperimentPart> experimentParts, HighLevelTaskFactory hltf, TaskThread tt) {
        super("Renew experiment until " + DateTimeFormatter.RFC_1123_DATE_TIME.format(ZonedDateTime.ofInstant(newExpirationDate, ZoneId.systemDefault())),
                experiment, hltf, tt);
        this.newExpirationDate = newExpirationDate;
        this.experimentParts = experimentParts;
    }

    @Override
    public ExperimentTaskStatus execute() throws Exception {
        if (experiment.getSlice().getExpirationDate() == null ||
                newExpirationDate.isAfter(experiment.getSlice().getExpirationDate())) {
            final RenewSliceState renewSliceState = new RenewSliceState();
            setAndRunState(renewSliceState);

            if (renewSliceState.getStatus() != ExperimentTaskStatus.SUCCESS) {
                throw new JFedHighLevelException("Could not renew slice, so renewing slivers will fail too!");
            }
        }

        RenewExperimentPartsState renewPartsState = new RenewExperimentPartsState();
        setAndRunState(renewPartsState);

        List<ExperimentPart> failedParts = renewPartsState.getThreads().stream()
                .filter(thread -> thread.getResult() != ExperimentTaskStatus.SUCCESS)
                .map(StatefulExperimentPartThread::getExperimentPart)
                .collect(Collectors.toList());

        if (!failedParts.isEmpty()) {
            updateMessage("Renewing failed at " + failedParts.stream().map(ExperimentPart::getName).collect(Collectors.joining(", ")));

        }

        return renewPartsState.getStatus();
    }

    private class RenewSliceState extends SingleCallState<RenewSliceTask> {

        protected RenewSliceState() {
            super("Renewing experiment container", RenewJob.this,
                    hltf.renewSlice(experiment.getSlice(), newExpirationDate));
        }
    }

    private class RenewExperimentPartsState extends ExperimentPartsThreadedState<RenewExperimentPartThread> {
        protected RenewExperimentPartsState() {
            super("Renewing resources at %d authorities", experiment, experimentParts);
        }

        @Override
        public Function<ExperimentPart, RenewExperimentPartThread> getExperimentPartToThreadFunction() {
            return RenewExperimentPartThread::new;
        }


    }

    private class RenewExperimentPartThread extends StatefulExperimentPartThread {

        public RenewExperimentPartThread(ExperimentPart experimentPart) {
            super(experimentPart);
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            RenewExperimentPartState renewExperimentPartState = new RenewExperimentPartState(experimentPart);
            setAndRunState(renewExperimentPartState);

            return renewExperimentPartState.getStatus();
        }
    }

    private class RenewExperimentPartState extends SingleCallState<RenewSliversTask> {

        protected RenewExperimentPartState(ExperimentPart experimentPart) {
            super(String.format("Renewing resources on %s", experimentPart.getName()),
                    RenewJob.this,
                    hltf.renewSliversAtAuthority(experiment.getSlice(), newExpirationDate, experimentPart.getSfaAuthority()));
        }

    }


}
