package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.ExperimentState;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.tasks.AggregatesForSliceTask;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.tasks.SliceManifestForAuthorityTask;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.util.GeniUrn;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.time.Instant;
import java.util.HashSet;
import java.util.Set;
import java.util.function.Function;

/**
 * User: twalcari
 * Date: 11/23/2015
 * Time: 16:47
 */
public class RestoreExperimentJob extends AbstractJob<Void> {
    private static final Logger LOG = LoggerFactory.getLogger(RestoreExperimentJob.class);


    private final AuthorityFinder authorityFinder;
    private final JobStateFactory jobStateFactory;

    public RestoreExperimentJob(Experiment experiment, HighLevelTaskFactory hltf, TaskThread tt,
                                AuthorityFinder authorityFinder, JobStateFactory jobStateFactory) {
        super("Restore Experiment", experiment, hltf, tt);
        this.authorityFinder = authorityFinder;
        this.jobStateFactory = jobStateFactory;
    }

    @Override
    public Void execute() throws Exception {
        //1. Check if we have sufficient information to start the restoration

        if (!hasSufficientSliceInformation()) {
            //first fetch the necessary info

            FetchSliceInfoState fetchSliceInfoState = new FetchSliceInfoState();
            setAndRunState(fetchSliceInfoState);

            assert fetchSliceInfoState.getTaskExecution().isCompleted();
            if (fetchSliceInfoState.getTaskExecution().getState() == TaskExecution.TaskState.SUCCESS) {
                assert (experiment.getSlice().getExpirationDate() != null);

                if (experiment.getSlice().getExpirationDate().isBefore(Instant.now())) {
                    experiment.setExperimentState(ExperimentState.EXPIRED_OR_EMPTY);

                    return null;
                }
            } else {
                //fetching slice info failed. We stop here!

                return null;
            }
        }

        assert hasSufficientSliceInformation();

        setAndRunState(new PopulateExperimentPartsState());

        //2. Now restore all parts of the experiment
        RestoreExperimentPartsState restoreExperimentPartsState = new RestoreExperimentPartsState();

        setAndRunState(restoreExperimentPartsState);

        experiment.setExperimentState(ExperimentState.CREATED);

        return null;
    }


    private boolean hasSufficientSliceInformation() {
        return (experiment.getSlice().getExpirationDate() != null);
    }


    private class FetchSliceInfoState extends SingleCallState<AggregatesForSliceTask> {
        public FetchSliceInfoState() {
            super("Requesting experiment information",
                    RestoreExperimentJob.this, hltf.getAggregatesForSlice(experiment.getSlice()));
        }
    }

    private class PopulateExperimentPartsState extends State {
        protected PopulateExperimentPartsState() {
            super("Processing experiment information");
        }

        @Override
        public ExperimentTaskStatus executeState() throws InterruptedException, JFedException {
            if (experiment.getExistingRequestRspecSource() == null) {
                LOG.info("No (additional) slivers to restore as there is no existingRequestRspecSource for this experiment!");
                return ExperimentTaskStatus.WARNING;
            }

            assert experiment.getExistingRequestRspecSource().getStringRspec().isWellFormed();

            Set<SfaAuthority> authorities = new HashSet<>();
            for (final GeniUrn componentManagerUrn : experiment.getExistingRequestRspecSource().getAllComponentManagerUrns()) {
                final SfaAuthority auth =
                        authorityFinder.findByUrn(componentManagerUrn, AuthorityFinder.Purpose.RECOVER);

                if (auth == null) {
                    LOG.error("ERROR: Component Manager Urn in Request RSpec unknown: \"" + componentManagerUrn + "\".");
                    throw new UnknownAuthorityException(componentManagerUrn);
                }

                if(auth.isFake())
                    continue;

                authorities.add(auth);
            }

            if (!authorities.isEmpty()) {
                for (SfaAuthority auth : authorities) {
                    if (experiment.getPart(auth) == null)
                        experiment.addExperimentPart(new ExperimentPart(experiment, auth));
                    else {
                        LOG.debug("An ExperimentPart already exists for authority {}", auth);
                    }
                }
            } else {

                LOG.debug("No additional authorities to contact");
            }

            return ExperimentTaskStatus.SUCCESS;
        }
    }

    private class RestoreExperimentPartsState extends ExperimentPartsThreadedState<RestoreExperimentPartThread> {
        public RestoreExperimentPartsState() {
            super(String.format("Requesting experiment information about %s from authorities ", experiment.getName()),
                    experiment);
        }

        @Override
        public Function<ExperimentPart, RestoreExperimentPartThread> getExperimentPartToThreadFunction() {
            return RestoreExperimentPartThread::new;
        }
    }

    private class RestoreExperimentPartThread extends StatefulExperimentPartThread {
        public RestoreExperimentPartThread(ExperimentPart experimentPart) {
            super(experimentPart);
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            if (experimentPart.getSfaAuthority().isFake()) {
                LOG.debug("Skipping restore of fake authority {}", experimentPart.getSfaAuthority().getName());
                return ExperimentTaskStatus.SUCCESS;
            }

            FetchSliceManifestForAuthorityState fetchManifestState =
                    new FetchSliceManifestForAuthorityState(experimentPart);
            setAndRunState(fetchManifestState);

            if (fetchManifestState.getTaskExecution().getState() != TaskExecution.TaskState.SUCCESS ||
                    experiment.getRequestedStartTime() != null && experiment.getRequestedStartTime().isAfter(Instant.now())) {
                // do nothing: it makes no sense to request status as the resources in the experiment part don't exist (yet).
                return ExperimentTaskStatus.WARNING;
            }

            GetStatusUntilReadyOrFailTaskState
                    untilReadyState =
                    jobStateFactory.createGetStatusUntilReadyOrFailTaskState(RestoreExperimentJob.this, experimentPart);
            setAndRunState(untilReadyState);

            return ExperimentTaskStatus.SUCCESS;
        }
    }


    private class FetchSliceManifestForAuthorityState extends SingleCallState<SliceManifestForAuthorityTask> {
        public FetchSliceManifestForAuthorityState(ExperimentPart experimentPart) {
            super(String.format("Requesting experiment information at %s", experimentPart.getName()),
                    RestoreExperimentJob.this,
                    hltf.getSliceManifest(experiment.getSlice(), experimentPart.getSfaAuthority()));
        }
    }

}
