package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.controller.JFedHighLevelException;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.model.EasyModel;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.highlevel.tasks.CreateSliverReservationTask;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.tasks.RequestSlaUuidTask;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.util.GeniUrn;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.time.Duration;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;
import java.util.stream.Collectors;

/**
 * This job is meant to start an experiment with the given details.
 * <p/>
 * It will create the slivers on all component managers that are mentioned in the request Rspec.
 * <p/>
 * First we try to make all the necessary reservations. Then we try to do all the necessary allocations.
 * <p/>
 * If one of these actions fail, a rollback is performed.
 * <p/>
 * Then we provision all the resources, wait for them to become ready, and end by doing a listResources
 * to get an up-to-date view of the resources supplied by the authority.
 * <p/>
 * User: twalcari
 * Date: 7/13/2015
 * Time: 16:48
 */
public class StartExperimentJob extends StartJob<Void> {
    private static final Logger LOG = LoggerFactory.getLogger(StartExperimentJob.class);

    private final AuthorityFinder authorityFinder;
    private final JobStateFactory jobStateFactory;
    private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;

    private List<UserSpec> userSpecs = null;

    private List<GeniUrn> failedResources = new ArrayList<>();


    StartExperimentJob(Experiment experiment,
                       HighLevelTaskFactory hltf, TaskThread tt,
                       AuthorityFinder authorityFinder,
                       AggregateManagerWrapperFactory aggregateManagerWrapperFactory,
                       GeniUserProvider geniUserProvider, EasyModel easyModel, JobStateFactory jobStateFactory) {
        super("Start Experiment " + experiment.getName(), experiment, hltf, tt, geniUserProvider, easyModel);

        this.authorityFinder = authorityFinder;
        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
        this.jobStateFactory = jobStateFactory;
    }

    @Override
    public Void execute() throws InterruptedException, JFedException {
        //-2. Do some preparations
        CreateUserSpecListState createUserSpecListState = jobStateFactory.createCreateUserSpecListState(experiment);
        setAndRunState(createUserSpecListState);

        userSpecs = createUserSpecListState.getUserSpecs();

        //-1. Request SLA UUID if necessary
        if (experiment.isRequestSla() && experiment.getSlice().getUuid() == null) {
            RequestSlaUuidState requestSlaUuidState = new RequestSlaUuidState();
            setAndRunState(requestSlaUuidState);
            if (requestSlaUuidState.getSlaUuid() != null) {
                experiment.getSlice().setUuid(requestSlaUuidState.getSlaUuid());
            } else {
                LOG.error("Did not receive a UUID from the SLA service!");
            }
        }

        //0. First gather all the authorities that we have to contact to start this
        setAndRunState(new PopulateExperimentPartsState());

        if (experiment.getParts().size() == 0) {
            LOG.warn("No ExperimentParts found to initialize!");
            return null;
        }

        StartSliversState sst = new StartSliversState();
        setAndRunState(sst);
        return null;
    }

    private class RequestSlaUuidState extends SingleCallState<RequestSlaUuidTask> {

        protected RequestSlaUuidState() {
            super("Requesting SLA UUID", StartExperimentJob.this, hltf.requestSlaUuid());
        }

        public String getSlaUuid() {
            return getTask().getSlaUuid();
        }
    }

    private class PopulateExperimentPartsState extends State {
        protected PopulateExperimentPartsState() {
            super("Processing experiment information");
        }

        @Override
        public ExperimentTaskStatus executeState() throws InterruptedException, JFedException {
            Set<SfaAuthority> authorities = new HashSet<>();
            for (final GeniUrn componentManagerUrn : experiment.getSlice().getRequestRspec().getAllComponentManagerUrns()) {
                final SfaAuthority auth =
                        authorityFinder.findByUrn(componentManagerUrn, AuthorityFinder.Purpose.CREATE_SLIVER);

                if (auth == null) {
                    LOG.error("ERROR: Component Manager Urn in Request RSpec unknown: \"" + componentManagerUrn + "\".");
                    throw new UnknownAuthorityException(componentManagerUrn);
                }

                if (auth.isFake())
                    continue;

                authorities.add(auth);
            }

            for (SfaAuthority auth : authorities) {
                experiment.addExperimentPart(new ExperimentPart(experiment, auth));
            }
            return ExperimentTaskStatus.SUCCESS;
        }

    }

    private class StartSliversState extends MultiThreadedState<InstantStartSliverThread> {
        private final CyclicBarrier barrier;


        private final Map<ExperimentPart, InstantStartSliverThread> experimentPartThreadMap = new HashMap<>();

        public StartSliversState() {
            super("Starting resources");
            this.barrier = new CyclicBarrier(experiment.getParts().size());

            for (ExperimentPart experimentPart : experiment.getParts().values()) {
                experimentPartThreadMap.put(experimentPart, new InstantStartSliverThread(experimentPart, barrier));
            }
        }

        @Override
        public Collection<InstantStartSliverThread> getThreads() {
            return experimentPartThreadMap.values();
        }
    }

    private class InstantStartSliverThread extends StatefulExperimentPartThread {
        private boolean hasResources = false;

        private final CyclicBarrier barrier;

        //gather info about authority
        private final boolean supportsReservation;
        private final boolean supportsAllocateProvision;

        public InstantStartSliverThread(ExperimentPart experimentPart, CyclicBarrier barrier) {
            super(experimentPart);
            this.barrier = barrier;
            this.supportsReservation = supportsReservation(experimentPart.getSfaAuthority());
            this.supportsAllocateProvision = supportsAllocateProvision(experimentPart.getSfaAuthority());
        }


        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            if (experimentPart.getSfaAuthority().isFake()) {
                //TODO: barriers kloppen hier niet
                LOG.debug("Skipping start ExperimentPart of fake authority {}", experimentPart.getSfaAuthority().getName());
                return ExperimentTaskStatus.SUCCESS;
            }

            final List<GeniUrn> createdSlivers = new ArrayList<>();
            //1a. Make reservation if authority supports it
            if (supportsReservation) {
                this.experimentPart.setState(InternalState.RESERVING);

                Instant startTime = experiment.getRequestedStartTime() != null ?
                        experiment.getRequestedStartTime() :
                        Instant.now().truncatedTo(ChronoUnit.SECONDS);

                CreateSliverReservationState csrs = new CreateSliverReservationState(startTime);
                setAndRunState(csrs);

                if (!csrs.getTask().isSuccess()) {
                    this.experimentPart.setState(InternalState.FAILED);

                    failedResources.addAll(csrs.getTask().getFailedResources());

                    barrier.reset();
                    return ExperimentTaskStatus.FAILED; //no further cleanup to do in this thread
                }


                csrs.getTask().getCreatedSlivers().stream()
                        .map(Sliver::getUrn)
                        .collect(Collectors.toCollection(() -> createdSlivers));

                hasResources = true;
                this.experimentPart.setState(InternalState.RESERVED);
            } else if (experiment.getRequestedStartTime() != null) {
                //TODO make sure that GUI doesn't allow this situation to happen
                throw new JFedHighLevelException("JFed doesn't support allocating and provisioning non-reservable resources in the future!");

            }

            try {
                barrier.await();
            } catch (BrokenBarrierException e) {
                LOG.error(logMarker, "Got broken barrier after reservation-step. Cleaning-up and aborting.", e);
                setAndRunState(jobStateFactory.createStopExperimentAtAuthorityPartState(StartExperimentJob.this, experimentPart));

                return ExperimentTaskStatus.FAILED;
            }

            //1b. If start-time was explicitely set, then wait for that start-time to happen
            if (experiment.getRequestedStartTime() != null) {
                setAndRunState(new WaitForExperimentStartTime());

                try {
                    barrier.await();
                } catch (BrokenBarrierException e) {
                    //TODO What happens when user closes jFed? Does this also break the barrier?
                    // If so: we don't want to cleanup in this case!

                    LOG.error(logMarker, "Got broken barrier after waiting for experiment to start. Cleaning-up and aborting.", e);
                    setAndRunState(jobStateFactory.createStopExperimentAtAuthorityPartState(StartExperimentJob.this, experimentPart));

                    return ExperimentTaskStatus.FAILED;
                }
            }


            //2a. Do allocation if authority doesn't support reservation
            // If separate allocation/provision is not supported, then we do a CreateSliver immediately


            if (!supportsReservation) {
                AllocateSliversOnAuthorityState
                        allocateSliversOnAuthorityState =
                        jobStateFactory.createAllocateSliversOnAuthorityState(StartExperimentJob.this, experimentPart, userSpecs);
                setAndRunState(allocateSliversOnAuthorityState);

                if (!allocateSliversOnAuthorityState.isSuccess()) {
                    this.experimentPart.setState(InternalState.FAILED);
                    barrier.reset();
                    return ExperimentTaskStatus.FAILED; //no further cleanup to do in this thread
                } else {
                    hasResources = true;
                }
                createdSlivers.addAll(allocateSliversOnAuthorityState.getCreatedSlivers());
                hasResources = true;

                assert experimentPart.getState() == InternalState.ALLOCATED ||
                        experimentPart.getState() == InternalState.PROVISIONED;
            }


            //2b. Submit created slivers to SLA-service
            if (experiment.isRequestSla())
                submitTask(hltf.registerSliversToSLACollector(experiment.getSlice(),
                        experimentPart.getSfaAuthority(), createdSlivers));
            else
                LOG.debug("Skipping registration of slivers to SLA Collector, as there is no SLA UUID");

            try {
                barrier.await();
            } catch (BrokenBarrierException e) {
                LOG.error(logMarker, "Got broken barrier after allocation-step. Cleaning-up and aborting.", e);
                setAndRunState(jobStateFactory.createStopExperimentAtAuthorityPartState(StartExperimentJob.this, experimentPart));

                return ExperimentTaskStatus.FAILED;
            }

            //3. Provision the resources if the authority supports it.
            if (supportsAllocateProvision) {
                final ProvisionSliversOnAuthorityState provisionSliversOnAuthorityState =
                        jobStateFactory.createProvisionSliversOnAuthorityState(StartExperimentJob.this, experimentPart, userSpecs);
                setAndRunState(provisionSliversOnAuthorityState);

                if (provisionSliversOnAuthorityState.getTaskExecution().getState() != TaskExecution.TaskState.SUCCESS) {
                    assert provisionSliversOnAuthorityState.getTaskExecution().getState() == TaskExecution.TaskState.FAILED ||
                            provisionSliversOnAuthorityState.getTaskExecution().getState() == TaskExecution.TaskState.CANCELLED
                            :
                            "Unexpected state for ProvisionSliverTask-execution : " + provisionSliversOnAuthorityState.getTaskExecution().getState();

                    this.experimentPart.setState(InternalState.FAILED);

                    LOG.error(logMarker, "Could not provision resources. Cleaning up and aborting.");

                    setAndRunState(jobStateFactory.createStopExperimentAtAuthorityPartState(StartExperimentJob.this, experimentPart));
                    return ExperimentTaskStatus.FAILED;
                }
                this.experimentPart.setState(InternalState.PROVISIONED);

                //TODO: register additional slivers to SLA Collector if there are any
            }

            setAndRunState(jobStateFactory.createGetStatusUntilReadyOrFailTaskState(StartExperimentJob.this, experimentPart));
            this.experimentPart.setState(InternalState.READY);

            return ExperimentTaskStatus.SUCCESS;
        }

        private class CreateSliverReservationState extends SingleCallState<CreateSliverReservationTask> {

            protected CreateSliverReservationState(Instant startTime) {
                super(String.format("Reserving resources on %s", experimentPart.getName()),
                        StartExperimentJob.this,
                        hltf.createSliverReservation(experiment.getSlice(),
                                experimentPart.getSfaAuthority(),
                                startTime,
                                experiment.getRequestedEndTime(),
                                userSpecs));
            }
        }


        private class WaitForExperimentStartTime extends be.iminds.ilabt.jfed.highlevel.jobs.State {

            protected WaitForExperimentStartTime() {
                super("Waiting for experiment reservation start time");
            }

            @Override
            public ExperimentTaskStatus executeState() throws InterruptedException, JFedException {
                assert experiment.getRequestedStartTime() != null;

                Duration durationToWait = Duration.between(Instant.now(), experiment.getRequestedStartTime());

                if (durationToWait.isNegative() || durationToWait.isZero())
                    return ExperimentTaskStatus.SUCCESS; //no need to wait

                Thread.sleep(durationToWait.toMillis());

                return ExperimentTaskStatus.SUCCESS;
            }
        }


    }

    private boolean supportsAllocateProvision(SfaAuthority auth) {
        AggregateManagerWrapper amWrapper =
                aggregateManagerWrapperFactory.getAggregateManagerWrapper(auth);

        return amWrapper.hasSeperateAllocateAndProvision();
    }

    public List<GeniUrn> getFailedResources() {
        return failedResources;
    }
}
