package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.call_log_output.HtmlLogOutput;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.jobs.report.JobReportWrappingLogger;
import be.iminds.ilabt.jfed.highlevel.jobs.report.StitchingJobReport;
import be.iminds.ilabt.jfed.highlevel.tasks.GetAdvertisementTask;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.tasks.SCSRequestTask;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractGeniAggregateManager;
import be.iminds.ilabt.jfed.lowlevel.api.StitchingComputationService;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.stitching.StitchingDirector;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.StringRspec;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.function.Function;

/**
 * User: twalcari
 * Date: 12/2/2015
 * Time: 16:43
 */
public class StartStitchedExperimentJob extends AbstractJob<Void> {
    private static final org.slf4j.Logger ACTUAL_LOG = LoggerFactory.getLogger(StartStitchedExperimentJob.class);
    private static final int TIME_BETWEEN_ALLOCATE_RETRIES_MS = 10_000;
    private static final int MAX_ALLOCATE_STITCH_ACTION_TRIES = 3;

    //this variable is typically set for debugging purposes
    private Integer waitForReadyDelay = null;

    private final SfaAuthority scsAuthority;
    private final JobStateFactory jobStateFactory;

    private final AuthorityListModel authorityListModel;
    private final AuthorityFinder authorityFinder;
    private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;

    private final StitchingDirector director;
    private final StitchingJobReport stitchingJobReport;
    private final JobReportWrappingLogger log;
    private final JobReportWrappingLogger stitchingDirectorLog;


    private List<UserSpec> userSpecs = null;
    private Boolean success = null;

    public StartStitchedExperimentJob(Experiment experiment, SfaAuthority scsAuthority, HighLevelTaskFactory hltf,
                                      TaskThread tt, JobStateFactory jobStateFactory,
                                      AuthorityListModel authorityListModel, AuthorityFinder authorityFinder,
                                      AggregateManagerWrapperFactory aggregateManagerWrapperFactory) {
        super("Start Experiment " + experiment.getName(), experiment, hltf, tt);
        this.scsAuthority = scsAuthority;
        this.jobStateFactory = jobStateFactory;
        this.authorityListModel = authorityListModel;
        this.authorityFinder = authorityFinder;
        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;

        assert experiment.getRequestedEndTime() != null;


        //do some initial setup for advanced logging

        this.stitchingJobReport = new StitchingJobReport(this);
        this.log = new JobReportWrappingLogger(ACTUAL_LOG, stitchingJobReport);
        this.stitchingDirectorLog = new JobReportWrappingLogger(StitchingDirector.ACTUAL_LOG, stitchingJobReport);

        this.director = new StitchingDirector(authorityListModel, authorityFinder,
                Date.from(experiment.getRequestedEndTime()), stitchingDirectorLog);
    }

    @Override
    public Void execute() throws Exception {
        //-2. Do some preparations
        CreateUserSpecListState createUserSpecListState = jobStateFactory.createCreateUserSpecListState(experiment);
        setAndRunState(createUserSpecListState);

        userSpecs = createUserSpecListState.getUserSpecs();

        log.debug("StartStitchedExperimentJob started");

        //make sure we have a valid experiment
        assert experiment.getSlice() != null;
        assert experiment.getRequestedEndTime() != null;


        SCSRequestState scsRequestState = new SCSRequestState();
        setAndRunState(scsRequestState);

        if (scsRequestState.getStatus() != ExperimentTaskStatus.SUCCESS) {
            log.error("SCS call failed", scsRequestState.getTaskExecution().getException());
            success = false;
            return null; //we give up
        }

        StitchingComputationService.ComputePathResult result = scsRequestState.getTask().getResult();
        assert result != null;

        //Update the slice request rspec with the enriched request-rspec from the SCS
        experiment.getSlice().setRequestRspec(new RequestRspecSource(new StringRspec(result.getServiceRspec()), ModelRspecType.FX));
        director.setComputePathResult(result);


        setAndRunState(new PopulateExperimentPartsState());


        boolean permanentFailure = false;
        boolean temporaryFailure = false;

        do {
            temporaryFailure = false;

            final AllocatePartsState allocatePartsState = new AllocatePartsState();
            setAndRunState(allocatePartsState);

            if (allocatePartsState.isPermanentFailure()) {
                permanentFailure = true;
                log.info("AllocatePartsState reported permanent failure");
                continue;
            }

            final ProvisionPartsState provisionPartsState = new ProvisionPartsState();
            setAndRunState(provisionPartsState);
            if (provisionPartsState.isPermanentFailure()) {
                permanentFailure = true;
                log.info("ProvisionPartsState reported permanent failure");
                continue;
            } else if (provisionPartsState.isTemporaryFailure()) {
                temporaryFailure = true;
                log.info("ProvisionPartsState reported temporary failure");
                continue;
            }

            final WaitForPartsReadyState waitForPartsReadyState = new WaitForPartsReadyState();
            setAndRunState(waitForPartsReadyState);

            if (waitForPartsReadyState.isPermanentFailure()) {
                permanentFailure = true;
                log.info("WaitForPartsReadyState reported permanent failure");
            } else if (waitForPartsReadyState.isTemporaryFailure()) {
                temporaryFailure = true;
                log.info("WaitForPartsReadyState reported temporary failure");
            }
        } while (!permanentFailure && temporaryFailure);

        if (permanentFailure) {
            DestroyExperimentPartsState destroyState = new DestroyExperimentPartsState();
            setAndRunState(destroyState);

            success = false;
        } else
            success = true;


        return null;
    }

    public Boolean isSuccess() {
        return success;
    }

    public StitchingJobReport getStitchingJobReport() {
        return stitchingJobReport;
    }

    public class SCSRequestState extends SingleCallState<SCSRequestTask> {
        //TODO: check mixed case of stitched and non-stitched authorities. Does SCS return non-stitched nodes/links in it's new RSpec?


        protected SCSRequestState() {
            super("Request stitched link setup-instructions",
                    StartStitchedExperimentJob.this,
                    hltf.createSCSRequestTask(experiment.getSlice(), experiment.getNewRequestRspecSource(),
                            experiment.getRequestedEndTime(), scsAuthority));
        }

        @Override
        public ExperimentTaskStatus executeState() throws InterruptedException {
            super.executeState();

            return getTask().getResult() != null ? ExperimentTaskStatus.SUCCESS : ExperimentTaskStatus.FAILED;
        }
    }

    private class PopulateExperimentPartsState extends State {
        protected PopulateExperimentPartsState() {
            super("Processing experiment information");
        }

        @Override
        public ExperimentTaskStatus executeState() throws InterruptedException, JFedException {
            for (SfaAuthority auth : director.getAllAuthorities()) {
                experiment.addExperimentPart(new ExperimentPart(experiment, auth));
            }
            return ExperimentTaskStatus.SUCCESS;
        }

    }

    public class AllocatePartsState extends State {
        private final Map<ExperimentPart, AllocatePartThread> threadMap = new HashMap<>();

        protected AllocatePartsState() {
            super("Allocate resources with StitchingDirector");
        }

        @Override
        protected ExperimentTaskStatus executeState() throws InterruptedException, JFedException {
            for (ExperimentPart experimentPart : experiment.getParts().values()) {
                AllocatePartThread allocatePartThread = new AllocatePartThread(experimentPart);
                threadMap.put(experimentPart, allocatePartThread);
                allocatePartThread.start();
            }


            while (hasDirectorAnyActionsLeft()) {
                List<StitchingDirector.StitchAction> stitchActions;
                synchronized (director) {
                    stitchActions = director.getStitchActions();
                }

                if (!stitchActions.isEmpty()) {

                    //  assert !stitchActions.isEmpty() :
                    //          "We expect to have stitch-actions, as director.arAnyActionsLeft()==true";


                    for (StitchingDirector.StitchAction stitchAction : stitchActions) {
                        ExperimentPart part = experiment.getPart(stitchAction.getAuthority());
                        threadMap.get(part).submitStitchAction(new AllocateAction(stitchAction));

                    }
                } else {
                    log.trace("No stitchActions available for the moment, but director says it's not finished yet.");
                }

                Thread.sleep(250);
            }
            threadMap.values().forEach(apt -> apt.submitStitchAction(AllocateAction.createFinishedAction()));

            return director.isGaveUp() ? ExperimentTaskStatus.FAILED : ExperimentTaskStatus.SUCCESS;
        }

        protected boolean hasDirectorAnyActionsLeft() {
            synchronized (director) {
                return director.areAnyActionsLeft();
            }
        }

        public boolean isPermanentFailure() {
            return director.isGaveUp();
        }
    }

    private static class AllocateAction {
        private final boolean finished;

        private final StitchingDirector.StitchAction stitchAction;


        public AllocateAction(StitchingDirector.StitchAction stitchAction) {
            this.stitchAction = stitchAction;
            this.finished = false;
        }

        private AllocateAction() {
            this.stitchAction = null;
            this.finished = true;
        }

        public static AllocateAction createFinishedAction() {
            return new AllocateAction();
        }

        public boolean isFinished() {
            return finished;
        }

        public StitchingDirector.StitchAction getStitchAction() {
            return stitchAction;
        }

        @Override
        public String toString() {
            return "AllocateAction{" +
                    "finished=" + finished +
                    ", stitchAction=" + stitchAction +
                    '}';
        }
    }

    private class AllocatePartThread extends StatefulExperimentPartThread {
        private final Logger log = StartStitchedExperimentJob.this.log;  //TODO: split into separate logger
        private final BlockingQueue<AllocateAction> queue = new LinkedBlockingQueue<>();

        private int allocateStitchActionRetryCount = 0;

        public AllocatePartThread(ExperimentPart experimentPart) {
            super(experimentPart);
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            while (true) {
                WaitForNextAllocateActionState waitForNextAllocateActionState = new WaitForNextAllocateActionState();
                setAndRunState(waitForNextAllocateActionState);

                AllocateAction allocateAction = waitForNextAllocateActionState.getAllocateAction();

                if (allocateAction.isFinished())
                    return ExperimentTaskStatus.SUCCESS;

                StitchingDirector.StitchAction stitchAction = allocateAction.getStitchAction();

                switch (stitchAction.getType()) {
                    case ALLOCATE:
                        processAllocateStitchAction(stitchAction);
                        break;
                    case DELETE:
                        processDeleteStitchAction();
                        break;
                    case LISTRESOURCES:
                        processListResourcesStitchAction();

                        break;
                    default:
                        throw new IllegalArgumentException("Unexpected StitchAction: " + stitchAction);
                }
            }

        }

        private void processListResourcesStitchAction() throws InterruptedException {
            GetAdvertisementState getAdvertisementState = new GetAdvertisementState();
            try {
                setAndRunState(getAdvertisementState);
            } catch (JFedException e) {
                log.error("Error while performing ListResources on {}", experimentPart.getSfaAuthority(), e);
            }

            synchronized (director) {
                director.processListResourcesResult(experimentPart.getSfaAuthority(),
                        getAdvertisementState.getTask().getAdvertisementRspecString());
            }
        }

        private void processDeleteStitchAction() throws InterruptedException {
            StopExperimentAtAuthorityPartState stopExperimentAtAuthorityPartState =
                    jobStateFactory.createStopExperimentAtAuthorityPartState(StartStitchedExperimentJob.this, experimentPart);
            try {
                setAndRunState(stopExperimentAtAuthorityPartState);
            } catch (JFedException e) {
                log.error("Error while deleting resources from experiment", e);
            } finally {
                synchronized (director) {
                    director.processDeleteResult(experimentPart.getSfaAuthority());
                }
            }
        }

        private void processAllocateStitchAction(StitchingDirector.StitchAction stitchAction) throws InterruptedException {
            RequestRspecSource overridingRequestRspecSource =
                    new RequestRspecSource(new StringRspec(stitchAction.getRequestRspec()));

            final AllocateSliversOnAuthorityState allocateSliversOnAuthorityState =
                    jobStateFactory.createAllocateSliversOnAuthorityState(StartStitchedExperimentJob.this, experimentPart, userSpecs, overridingRequestRspecSource);


            AbstractGeniAggregateManager.AggregateManagerReply reply = null;

            try {
                setAndRunState(allocateSliversOnAuthorityState);

                if (allocateSliversOnAuthorityState.isSuccess()) {
                    reply = allocateSliversOnAuthorityState.getRawReply();

                } else if (allocateSliversOnAuthorityState.getException() instanceof AggregateManagerWrapper.VlanUnavailableException ||
                        allocateSliversOnAuthorityState.getException() instanceof AggregateManagerWrapper.AnyVlanUnsupportedException) {
                    JFedException e = (JFedException) allocateSliversOnAuthorityState.getException();
                    assert e.getSfaReply() == null || e.getSfaReply() instanceof AbstractGeniAggregateManager.AggregateManagerReply :
                            "Unexpected reply type: " + e.getSfaReply().getClass().getName();
                    //ignore
                    reply = (AbstractGeniAggregateManager.AggregateManagerReply) e.getSfaReply();
                } else if (allocateSliversOnAuthorityState.getException() instanceof JFedException) {

                    JFedException e = (JFedException) allocateSliversOnAuthorityState.getException();

                    if (e.getSfaReply() != null && e.getSfaReply() instanceof AbstractGeniAggregateManager.AggregateManagerReply) {
                        reply = (AbstractGeniAggregateManager.AggregateManagerReply) e.getSfaReply();
                    } else {
                        if (StitchingDirector.isUnrecoverableError(e) == Boolean.TRUE) {
                            log.error("Giving up as {} reported an unrecoverable error {}", experimentPart.getName(), e.getMessage());
                            synchronized (director) {
                                director.reportGiveup();
                            }
                        } else {
                            retryAllocate(stitchAction);
                            return; //we don't want to process this allocate-try any further
                        }
                    }
                } else {
                    log.error("Unexpected state: unknown error while allocating", allocateSliversOnAuthorityState.getException());
                }
            } catch (JFedException ex) {
                log.error("UNEXPECTED error while allocating", ex);
            }

            log.trace("   Processing result of CreateSliver call for " + experimentPart.getName());
            StitchingDirector.AllocateVerdict verdict;
            synchronized (director) {
                verdict = director.processAllocateResult(stitchAction, reply);
            }

            if (verdict == StitchingDirector.AllocateVerdict.CALLER_DELAYED_RETRY) {
                retryAllocate(stitchAction);
            }
        }

        private void retryAllocate(StitchingDirector.StitchAction stitchAction) throws InterruptedException {
            if (allocateStitchActionRetryCount++ > MAX_ALLOCATE_STITCH_ACTION_TRIES) {
                log.error("Giving up stitching as number of allocation tries on {} exceeded maximum of {}", experimentPart.getName(), MAX_ALLOCATE_STITCH_ACTION_TRIES);
                synchronized (director) {
                    director.reportGiveup();
                }
                return;
            }

            log.info("retryAllocate: CreateSliver failed (allocateStitchActionRetryCount=" + allocateStitchActionRetryCount + "), " +
                    "will delete and try again in " + TIME_BETWEEN_ALLOCATE_RETRIES_MS + "ms " +
                    "now=" + HtmlLogOutput.dateToString(new Date()));

            Thread.sleep(TIME_BETWEEN_ALLOCATE_RETRIES_MS);
            synchronized (director) {
                director.reportReadyForRetry(experimentPart.getSfaAuthority());
            }
        }


        public void submitStitchAction(AllocateAction stitchAction) {
            boolean success = queue.offer(stitchAction);
            assert success : "Could not add StitchAction to ExperimentPart " + experimentPart.getName();
        }


        public class GetAdvertisementState extends SingleCallState<GetAdvertisementTask> {
            protected GetAdvertisementState() {
                super("Fetch advertisement from " + experiment.getName() + " for list of available VLAN's", StartStitchedExperimentJob.this,
                        hltf.getAvailableAdverisement(experimentPart.getSfaAuthority()));
            }
        }

        public class WaitForNextAllocateActionState extends be.iminds.ilabt.jfed.highlevel.jobs.State {

            private AllocateAction allocateAction;

            protected WaitForNextAllocateActionState() {
                super("Waiting for next StitchAction instruction");
            }

            @Override
            protected ExperimentTaskStatus executeState() throws InterruptedException, JFedException {
                allocateAction = queue.take();

                return ExperimentTaskStatus.SUCCESS;
            }

            public AllocateAction getAllocateAction() {
                return allocateAction;
            }
        }
    }


    public class ProvisionPartsState extends ExperimentPartsThreadedState<ProvisionPartThread> {

        protected ProvisionPartsState() {
            super("Provision all ExperimentParts", experiment);
        }

        @Override
        public Function<ExperimentPart, ProvisionPartThread> getExperimentPartToThreadFunction() {
            return ProvisionPartThread::new;
        }

        public boolean isPermanentFailure() {
            return false;
        }

        public boolean isTemporaryFailure() {
            return false;
        }
    }

    public class ProvisionPartThread extends StatefulExperimentPartThread {

        public ProvisionPartThread(ExperimentPart experimentPart) {
            super(experimentPart);
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            if (!supportsAllocateProvision(experimentPart.getSfaAuthority())) {
                return ExperimentTaskStatus.SUCCESS;
            }

            ProvisionSliversOnAuthorityState provisionSliversOnAuthorityState =
                    jobStateFactory.createProvisionSliversOnAuthorityState(StartStitchedExperimentJob.this, experimentPart, userSpecs);

            setAndRunState(provisionSliversOnAuthorityState);

            return provisionSliversOnAuthorityState.getStatus();

        }

        private boolean supportsAllocateProvision(SfaAuthority auth) {
            AggregateManagerWrapper amWrapper =
                    aggregateManagerWrapperFactory.getAggregateManagerWrapper(auth);

            return amWrapper.hasSeperateAllocateAndProvision();
        }

    }


    public class WaitForPartsReadyState extends ExperimentPartsThreadedState<WaitForPartReadyThread> {

        protected WaitForPartsReadyState() {
            super("Waiting for resources to become ready", experiment);
        }

        @Override
        public Function<ExperimentPart, WaitForPartReadyThread> getExperimentPartToThreadFunction() {
            return WaitForPartReadyThread::new;
        }

        public boolean isPermanentFailure() {
            return false;
        }

        public boolean isTemporaryFailure() {
            return false;
        }
    }

    public class WaitForPartReadyThread extends StatefulExperimentPartThread {

        public WaitForPartReadyThread(ExperimentPart experimentPart) {
            super(experimentPart);
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            GetStatusUntilReadyOrFailTaskState statusState =
                    jobStateFactory.createGetStatusUntilReadyOrFailTaskState(StartStitchedExperimentJob.this, experimentPart);

            if (waitForReadyDelay != null)
                statusState.setWaitForReadyDelay(waitForReadyDelay);

            setAndRunState(statusState);
            return statusState.getStatus();
        }
    }

    public class DestroyExperimentPartsState extends ExperimentPartsThreadedState<DestroyExperimentPartThread> {

        protected DestroyExperimentPartsState() {
            super("Delete all resources", experiment);
        }

        @Override
        public Function<ExperimentPart, DestroyExperimentPartThread> getExperimentPartToThreadFunction() {
            return DestroyExperimentPartThread::new;
        }
    }

    public class DestroyExperimentPartThread extends StatefulExperimentPartThread {

        public DestroyExperimentPartThread(ExperimentPart experimentPart) {
            super(experimentPart);
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            StopExperimentAtAuthorityPartState stopState =
                    jobStateFactory.createStopExperimentAtAuthorityPartState(StartStitchedExperimentJob.this, experimentPart);
            setAndRunState(stopState);
            return stopState.getStatus();
        }
    }

    public void setWaitForReadyDelay(Integer waitForReadyDelay) {
        this.waitForReadyDelay = waitForReadyDelay;
    }
}
