package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.ExperimentState;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.controller.JFedHighLevelException;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.tasks.RegisterStopToSLACollectorTask;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.time.Instant;
import java.util.Collection;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 12/1/2015
 * Time: 17:03
 */

//TODO: replace this by job that starts StopExperimentAtAuthorityTask for each ExperimentPart
public class StopExperimentJob extends AbstractJob<Void> {
    private static final Logger LOG = LoggerFactory.getLogger(StopExperimentJob.class);
    private final Collection<ExperimentPart> experimentParts;
    private final JobStateFactory jobStateFactory;

    public StopExperimentJob(Experiment experiment, HighLevelTaskFactory hltf, TaskThread tt, JobStateFactory jobStateFactory) {
        super("Delete all active resources from experiment '" + experiment.getName() + "'", experiment, hltf, tt);
        this.jobStateFactory = jobStateFactory;
        this.experimentParts = null;
    }

    public StopExperimentJob(Experiment experiment, Collection<ExperimentPart> experimentParts, HighLevelTaskFactory hltf, TaskThread tt, JobStateFactory jobStateFactory) {
        super("Delete active resources from '" + experimentParts.stream().map(ExperimentPart::getName).collect(Collectors.joining(", ")) + "'",
                experiment, hltf, tt);
        this.experimentParts = experimentParts;
        this.jobStateFactory = jobStateFactory;
    }


    @Override
    public Void execute() throws Exception {
        assert experiment.getSlice() != null;
        if (experiment.getSlice().isExpired()) {
            throw new JFedHighLevelException("Slice has expired.");
        }
        if (!experiment.getSlice().hasCredential()) {
            throw new JFedHighLevelException("No credentials for slice. Has it expired? slice expiration date = "
                    + experiment.getSlice().getExpirationDate() + " now=" + Instant.now());
        }

        if (experimentParts == null)
            experiment.setExperimentState(ExperimentState.DELETING);

        StopExperimentAtExperimentPartsState stopExperimentAtExperimentPartsState =
                new StopExperimentAtExperimentPartsState();
        setAndRunState(stopExperimentAtExperimentPartsState);


        if (experimentParts == null) {
            if (stopExperimentAtExperimentPartsState.getStatus() == ExperimentTaskStatus.SUCCESS) {
                experiment.setExperimentState(ExperimentState.EXPIRED_OR_EMPTY);
            } else {
                LOG.warn("Error while stopping experiment: state did not return SUCCESS");
            }
        } else {
            //check if there are still active resources
            boolean hasActiveResources = false;

            for (StatusDetails.SliverStatus sliverStatus : experiment.getSlice().getStatus().getStatusBySliverUrn().values()) {
                if (sliverStatus == StatusDetails.SliverStatus.READY || sliverStatus == StatusDetails.SliverStatus.CHANGING)
                    hasActiveResources = true;
            }

            for (StatusDetails.SliverStatus sliverStatus : experiment.getSlice().getStatus().getStatusByComponentUrn().values()) {
                if (sliverStatus == StatusDetails.SliverStatus.READY || sliverStatus == StatusDetails.SliverStatus.CHANGING)
                    hasActiveResources = true;
            }

            if (!hasActiveResources) {
                experiment.setExperimentState(ExperimentState.EXPIRED_OR_EMPTY);
                experiment.getSlice().setHasActiveSlivers(false);
            } else {
                assert (experiment.getExperimentState() == ExperimentState.DELETING);
            }
        }


        return null;
    }

    public class StopExperimentAtExperimentPartsState extends ExperimentPartsThreadedState<StopExperimentAtAuthorityPartThread> {
        protected StopExperimentAtExperimentPartsState() {
            super("Delete all active resources from experiment '" + experiment.getName() + "'", experiment, experimentParts);
        }

        @Override
        public Function<ExperimentPart, StopExperimentAtAuthorityPartThread> getExperimentPartToThreadFunction() {
            return StopExperimentAtAuthorityPartThread::new;
        }
    }

    private class StopExperimentAtAuthorityPartThread extends StatefulExperimentPartThread {

        public StopExperimentAtAuthorityPartThread(ExperimentPart experimentPart) {
            super(experimentPart);
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            experimentPart.setState(InternalState.DELETING);

            final StopExperimentAtAuthorityPartState stopPartState =
                    jobStateFactory.createStopExperimentAtAuthorityPartState(StopExperimentJob.this, experimentPart);
            setAndRunState(stopPartState);

            if (experiment.isRequestSla()) {
                if (experiment.getSlice().getUuid() != null) {
                    setAndRunState(new RegisterStopToSlaCollectorState(experimentPart));
                } else {
                    LOG.warn("Could not register stop of experiment {} at authority {} to SLA Collector because no UUID is available",
                            experiment.getName(), experimentPart.getName());
                }

            }

            if (stopPartState.getStatus() == ExperimentTaskStatus.SUCCESS)
                experimentPart.setState(InternalState.DELETED);
            else
                experimentPart.setState(InternalState.FAILED);

            return stopPartState.getStatus();
        }
    }

    private class RegisterStopToSlaCollectorState extends SingleCallState<RegisterStopToSLACollectorTask> {

        protected RegisterStopToSlaCollectorState(ExperimentPart experimentPart) {
            super("Registering termination of resources to SLA Collector", StopExperimentJob.this,
                    hltf.registerStopToSLACollector(experiment.getSlice(), experimentPart.getSfaAuthority()));
        }
    }

}
