package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.ExperimentState;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.tasks.SliceStatusAtAuthorityTask;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.function.Function;

/**
 * User: twalcari
 * Date: 11/25/2015
 * Time: 17:40
 */
public class UpdateExperimentJob extends AbstractJob<Void> {
    private static final Logger LOG = LoggerFactory.getLogger(UpdateExperimentJob.class);


    public UpdateExperimentJob(Experiment experiment, HighLevelTaskFactory hltf, TaskThread tt) {
        super("Check expiration of " + experiment.getName(), experiment, hltf, tt);
    }

    @Override
    public Void execute() throws Exception {
        assert experiment.getSlice() != null;
        assert experiment.getSlice().getExpirationDate() != null;

        //first check if expiration time has changed without we knowing it.
        setAndRunState(new UpdateExperimentState());

        //this will be the case when this job was started because one or more resources have reached their expiration time
        if (experiment.getExperimentState() == ExperimentState.EXPIRING) {
            boolean allResourcesExpired = true;
            boolean hasChangingExperimentParts = false;


            for (ExperimentPart part : experiment.getParts().values()) {
                if (part.getState() != InternalState.UNALLOCATED
                        && part.getState() != InternalState.DELETED
                        && part.getState() != InternalState.FAILED)
                    allResourcesExpired = false;

                if (part.getState() == InternalState.CHANGING)
                    hasChangingExperimentParts = true;
            }

            for (Sliver sliver : experiment.getSlice().getSlivers()) {
                if (sliver.getStatus().getGlobalStatus() != StatusDetails.SliverStatus.UNALLOCATED &&
                        sliver.getStatus().getGlobalStatus() != StatusDetails.SliverStatus.FAIL)
                    allResourcesExpired = false;
            }

            if (allResourcesExpired) {
                experiment.setExperimentState(ExperimentState.EXPIRED_OR_EMPTY);
            } else if (!hasChangingExperimentParts) {
                //check if one or more experimentparts are changing
                experiment.setExperimentState(ExperimentState.READY);
            } else {
                //TODO: do we start updating the experimentpart that is changing here? Or is this done elsewere?
            }
        }

        return null;
    }

    private class UpdateExperimentState extends ExperimentPartsThreadedState<UpdateExperimentPartThread> {
        public UpdateExperimentState() {
            super("Updating experiment status", experiment);
        }

        @Override
        public Function<ExperimentPart, UpdateExperimentPartThread> getExperimentPartToThreadFunction() {
            return UpdateExperimentPartThread::new;
        }
    }

    private class UpdateExperimentPartThread extends StatefulExperimentPartThread {
        public UpdateExperimentPartThread(ExperimentPart experimentPart) {
            super(experimentPart);
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            assert experiment.getSlice() != null;

            setAndRunState(new UpdateExperimentPartState());

            //check if status of resources in ExperimentPart has changed
            assert experimentPart.getSlivers() != null;

            StatusDetails.SliverStatus combinedStatus =
                    experimentPart.getSlivers().stream()
                            .map(sliver -> sliver.getStatus().getGlobalStatus())
                            .reduce(StatusDetails.SliverStatus.READY, StatusDetails.SliverStatus::merge);

            experimentPart.setState(sliverStatusToInternalState(combinedStatus));

            return combinedStatus != StatusDetails.SliverStatus.FAIL ? ExperimentTaskStatus.SUCCESS
                    : ExperimentTaskStatus.FAILED;
        }


        private class UpdateExperimentPartState extends SingleCallState<SliceStatusAtAuthorityTask> {
            @SuppressWarnings("ConstantConditions")
            public UpdateExperimentPartState() {
                super(String.format("Updating experiment status on %s", experimentPart.getName()),
                        UpdateExperimentJob.this, hltf.getSliceStatus(experiment.getSlice(), experimentPart.getSfaAuthority()));

            }
        }

    }

    private static InternalState sliverStatusToInternalState(StatusDetails.SliverStatus sliverStatus) {
        switch (sliverStatus) {
            case READY:
                return InternalState.READY;
            case FAIL:
                return InternalState.FAILED;
            case CHANGING:
                return InternalState.CHANGING;
            case UNALLOCATED:
                return InternalState.UNALLOCATED;
            case UNINITIALISED:
            case UNKNOWN:
                return InternalState.UNKNOWN;
            default:
                throw new IllegalArgumentException("Unexpected SliverStatus: " + sliverStatus);
        }
    }
}
