package be.iminds.ilabt.jfed.highlevel.tasks;

import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.util.JavaFXLogger;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CountDownLatch;

/**
 * User: twalcari
 * Date: 6/17/2015
 * Time: 15:12
 */
public class SliceStatusTask extends Task {
    private static final Logger LOG = LoggerFactory.getLogger(SliceStatusTask.class);

    private final Slice slice;

    private final TaskThread tt;
    private final HighLevelTaskFactory highLevelTaskFactory;
    private final JavaFXLogger logger;
    private final UserAndSliceApiWrapper userAndSliceApiWrapper;


    public SliceStatusTask(Slice slice, TaskThread tt, HighLevelTaskFactory highLevelTaskFactory,
                           JavaFXLogger logger, UserAndSliceApiWrapper userAndSliceApiWrapper) {
        super("Status of Slice");
        this.slice = slice;
        this.tt = tt;
        this.highLevelTaskFactory = highLevelTaskFactory;
        this.logger = logger;
        this.userAndSliceApiWrapper = userAndSliceApiWrapper;
    }

    @Override
    public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {

        //create a copy to prevent deadlocks because of a changing number of authorities
        Set<SfaAuthority> manifestAuthorities = new HashSet<>(slice.getManifestComponentManagers());

        //with this latch, we wait for all child-calls to finish before returning ourselves.
        CountDownLatch callsDoneLatch = new CountDownLatch(manifestAuthorities.size());

        for (SfaAuthority auth : manifestAuthorities) {
            if (auth.hasAmUrl()) {
                tt.addTask(highLevelTaskFactory.getSliceStatus(slice, auth),
                        (task, subTaskEx, state) -> callsDoneLatch.countDown());

            } else {
                LOG.error("Tried to get Status for slice '{}' at authority {}, " +
                                "but that authority has no AM URL: will ignore slice on this authority.",
                        slice.getUrnString(), auth.getUrnString());
                callsDoneLatch.countDown();
            }
        }

        callsDoneLatch.await();

        //check if we now have an inconsistent slice expiration time
        assert slice.getExpirationDate() != null;
        if (slice.getSlivers().stream()
                .anyMatch(sliver -> sliver.getExpirationDate() != null &&
                        sliver.getExpirationDate().isAfter(slice.getExpirationDate()))) {
            userAndSliceApiWrapper.getSliceExpiration(
                    JavaFXLogger.wrap(logger, taskExecution), slice.getCredentials(), slice.getUrn());
        }


    }

    @Override
    public List<Task> initDependsOn() {
        return Collections.singletonList(highLevelTaskFactory.getSliceCredential(slice));
    }

    @Override
    public List<Task> initAlwaysDependsOn() {
        return Collections.singletonList(highLevelTaskFactory.getAggregatesForSlice(slice));
    }
}
